library(MASS)
library(nleqslv)
library(Matrix)
library(glm2)
library(numDeriv)
library(mvmeta)
library(Rcpp)
library(RcppArmadillo)

sourceCpp("./cpppls.cpp")

# Prepare integrand
normal = function(m,v){
	function(x){
		1/sqrt(2*pi*v^2)*exp(-(x-m)^2/2/v^2)
	}	
}

normalx = function(m,v){
	function(x){
		x/sqrt(2*pi*v^2)*exp(-(x-m)^2/2/v^2)
	}	
}

# Likeihood function for mu and sigma of distribution of X1
Likelihood = function(cutoff_bin,cutoff_ter,m_x,v_x,n){ 
	like = function(x){

		m = x[1]
		v = x[2]

		-1*(
		#likelihood for study 1-3 (i.e., studies with continuous values) (mean value)
		log(1/sqrt(2*pi*v^2/n[[1]])*exp(-(m_x[[1]]-m)^2/2/v^2/n[[1]]))+ 
		log(1/sqrt(2*pi*v^2/n[[2]])*exp(-(m_x[[2]]-m)^2/2/v^2/n[[2]]))+
		log(1/sqrt(2*pi*v^2/n[[3]])*exp(-(m_x[[3]]-m)^2/2/v^2/n[[3]]))+
		
		#likelihood for study 1-3 (sd value)
		log(1/(2/n[[1]]*v^2*gamma((n[[1]]-1)))*(v_x[[1]]^2/(2/n[[1]]*v^2))^((n[[1]]-1)/2-1)*exp(-v_x[[1]]^2/(2/n[[1]]*v^2)))+
		log(1/(2/n[[2]]*v^2*gamma((n[[2]]-1)))*(v_x[[2]]^2/(2/n[[1]]*v^2))^((n[[2]]-1)/2-1)*exp(-v_x[[2]]^2/(2/n[[2]]*v^2)))+
		log(1/(2/n[[3]]*v^2*gamma((n[[3]]-1)))*(v_x[[3]]^2/(2/n[[1]]*v^2))^((n[[3]]-1)/2-1)*exp(-v_x[[3]]^2/(2/n[[3]]*v^2)))+

		#likelihood for study 4-6 (i.e., studies with binary values) 
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[1])$value)*as.numeric(n[[4]][1])+log(1-integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[1])$value)*as.numeric(n[[4]][2])+
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[2])$value)*as.numeric(n[[5]][1])+log(1-integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[2])$value)*as.numeric(n[[5]][2])+
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[3])$value)*as.numeric(n[[6]][1])+log(1-integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_bin[3])$value)*as.numeric(n[[6]][2])+
		
		#likelihood for study 7-9 (i.e., studies with tertiary values)
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_ter[1,1])$value)*as.numeric(n[[7]][1])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[1,1],upper=cutoff_ter[1,2])$value)*as.numeric(n[[7]][2])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[1,2],upper=Inf)$value)*as.numeric(n[[7]][3])+
		
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_ter[2,1])$value)*as.numeric(n[[8]][1])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[2,1],upper=cutoff_ter[2,2])$value)*as.numeric(n[[8]][2])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[2,2],upper=Inf)$value)*as.numeric(n[[8]][3])+
		
		log(integrate(normal(m=m,v=v),lower=-Inf,upper=cutoff_ter[3,1])$value)*as.numeric(n[[9]][1])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[3,1],upper=cutoff_ter[3,2])$value)*as.numeric(n[[9]][2])+
		log(integrate(normal(m=m,v=v),lower=cutoff_ter[3,2],upper=Inf)$value)*as.numeric(n[[9]][3])
		)
	}
	return(like)
}

# This is f_0i,f_1i,f_2i in the study 4-6.
# These functions can be obtained by solving equation (3).
step1 = function(par.a, cutoff, sample, mu, rho, sx, sz){
	a0 = par.a[1]
	a1 = par.a[2]
	a2 = par.a[3]
	
	a = integrate(normal(m = mu[1], v = sx), lower = cutoff, upper = Inf)$value
	b = integrate(normalx(m = mu[1], v = sx), lower = cutoff, upper = Inf)$value
	c = mu[2]*a+rho*sqrt(sz^2/sx^2)*(b-mu[1]*a)
	 
	E_xz = rho*sqrt(sx^2*sz^2)+mu[1]*mu[2]
	E_z2 = sz^2+mu[2]^2

	par.b1 = (a1*(mu[1]*mu[2]-E_xz)-a1*(mu[1]*a-b)*(E_z2-mu[2]^2)/(c-a*mu[2]))/(a*mu[2]-c-a*(a-1)/(c-a*mu[2]))
	par.b2 = a2-(a1*(mu[1]*a-b)-par.b1*a*(a-1))/(c-a*mu[2])
	par.b0 = a0+a1*mu[1]-par.b1*a+(a2-par.b2)*mu[2]

	return(c(par.b0,par.b1,par.b2))
}

# This is f_0i,f_1i,f_2i in the study 7-9.
# These functions can be obtained by solving equation (3).
step2 = function(par.a, cutoff1, cutoff2, sample, mu, rho, sx, sz){
	a0 = par.a[1]
	a1 = par.a[2]
	a2 = par.a[3]

	a = integrate(normal(m = mu[1], v = sx), lower = cutoff1, upper = cutoff2)$value
	b = integrate(normal(m = mu[1], v = sx), lower = cutoff2, upper = Inf)$value
	c = integrate(normalx(m = mu[1], v = sx), lower = cutoff1, upper = cutoff2)$value
	d = mu[2]*a+rho*sqrt(sz^2/sx^2)*(c-mu[1]*a)
	e = integrate(normalx(m = mu[1], v = sx), lower = cutoff2, upper = Inf)$value
	f = mu[2]*b+rho*sqrt(sz^2/sx^2)*(e-mu[1]*b)

	k1 = a+2*b
	k2 = d+2*f
	k3 = c+2*e	
	E_xz = rho*sqrt(sx^2*sz^2)+mu[1]*mu[2]
	E_z2 = sz^2+mu[2]^2
	
	par.c1 = (a1*(mu[1]*mu[2]-E_xz)-(a1*(k3-k1*mu[1]))/(mu[2]*k1-k2)*(E_z2-mu[2]^2))/(k1*mu[2]-k2+((k1-1)*a+2*(k1-2)*b)/(mu[2]*k1-k2)*(E_z2-mu[2]^2)) 
	par.c2 = a2 - (a1*(k3-mu[1]*k1)+par.c1*((k1-1)*a+2*(k1-2)*b))/(mu[2]*k1-k2) 
	par.c0 = a0+a1*mu[1]-par.c1*k1+(a2-par.c2)*mu[2]

	return(c(par.c0,par.c1,par.c2))

}


# this is function for calculating the first component of likeliihood, defined the function "ll.meta", under the case of study 4-6 
term1 = function(S, sum, dist, b, i, Sigma){
	df.dx.b = jacobian(function(x){step1(x, cutoff = cutoff_bin[i-3],mu = c(dist[[i]][[1]][1],dist[[i]][[1]][2]),rho=rho, sx = sqrt(dist[[i]][[2]][1,1]), sz = sqrt(dist[[i]][[2]][2,2]))}, x = b)
	term_tmp2 = t(df.dx.b) %*% solve(S[[i]]) %*% df.dx.b + solve(Sigma)	
	return(log(det(term_tmp2)))	
}

# this is function for calculating the first component of likeliihood, defined the function "ll.meta", under the case of study 7-9
term2 = function(S, sum, dist, b, i, Sigma){
	df.dx.b = jacobian(function(x){step2(x, cutoff1 = cutoff_ter[i-6,1], cutoff2 = cutoff_ter[i-6,2],mu = c(dist[[i]][[1]][1],dist[[i]][[1]][2]), rho = rho, sx = sqrt(dist[[i]][[2]][1,1]), sz = sqrt(dist[[i]][[2]][2,2]))}, x = b)
	term_tmp3 = t(df.dx.b) %*% solve(S[[i]]) %*% df.dx.b + solve(Sigma)
	return(log(det(term_tmp3)))
}

# This likelihood for random-effect meta-analysis is obtained based on Pinheiro and Bates (2000), Chapter 7.
# Pinheiro, J. C. and D. M. Bates (2000). Mixed-Effects Models in S and S-PLUS. New York, SpringerVerlag.
ll.meta = function(pars,coef,dist,S){
	all0 = pars[1]
	all1 = pars[2]
	all2 = pars[3]

	Xdiag<-matrix(c(
		pars[4],0,0,
		0,pars[5],0,
		0,0,pars[6]
	),nrow=3)
	Rmat<-matrix(c(
		1, pars[7],pars[8],
		pars[7],1,pars[9],
		pars[8],pars[9],1,
	),nrow=3)
	Sigma = Xdiag%*%Rmat%*%Xdiag

	#Make sure that Sigma is positive semi-definite
	Sigma = as.matrix(nearPD(Sigma)$mat)

	tau = list(9)
	for(i in 1:9){
		if(i %in% c(1,2,3)){
			tau[[i]]=optim(c(rep(0,3)),function(x){pls_i(coef = coef[i,],
				Sigma = Sigma,
				tau = x,
				as = x,
				Sigma_with = S[[i]],
				all = c(all0,all1,all2)
				)},control=list(maxit=10000),method="BFGS")		
		}else if(i %in% c(4,5,6)){
			tau[[i]]=optim(c(rep(0,3)),function(x){pls_i(coef = coef[i,],
				Sigma = Sigma,
				tau = step1(x, cutoff = cutoff_bin[i-3],mu = c(dist[[i]][[1]][1],dist[[i]][[1]][2]),rho=rho, sx = sqrt(dist[[i]][[2]][1,1]), sz = sqrt(dist[[i]][[2]][2,2])),
				as = x,
				Sigma_with = S[[i]],
				all = c(all0,all1,all2)
				)},control=list(maxit=10000),method="BFGS")		
		}else{
			tau[[i]]=optim(c(rep(0,3)),function(x){pls_i(coef = coef[i,],
				Sigma = Sigma,
				tau = step2(x, cutoff1 = cutoff_ter[i-6,1], cutoff2 = cutoff_ter[i-6,2],mu = c(dist[[i]][[1]][1],dist[[i]][[1]][2]), rho = rho, sx = sqrt(dist[[i]][[2]][1,1]), sz = sqrt(dist[[i]][[2]][2,2])),
				as = x,
				Sigma_with = S[[i]],
				all = c(all0,all1,all2)
				)},control=list(maxit=10000),method="BFGS")	
		}
	}


	# first component of log likelihood
	first = 0
	for(i in 1:3){
		first = first + log(det(solve(S[[i]])+solve(Sigma)))
	}
	for(i in 4:6){	
		first = first + term1(S, sum, dist, tau[[i]]$par, i, Sigma = Sigma)
	}	
	for(i in 7:9){		
		first = first + term2(S, sum, dist, tau[[i]]$par, i, Sigma = Sigma)
	}

	# second component of log likelihood
	second = sum(sapply(tau,function(x){x[[2]]}))

	# third component of log likelihood
	third = 0
	for(i in 1:9){
		third = third + log(det(S[[i]]))
	}

	# calculate the log likeilihood for meta-analysis based on the expression (4), (5) and (6)
	logl = -1*(4.5*log(det(chol(solve(Sigma))))-0.5*(first+second+third))

	return(logl)
}


# coef: matrix of reported coefficients in the study 1-9 (1 study 1 row)
# S: list of reported within-study covariance matrices, which is same as what mvmeta package requires. 
# m_x: list of reported means of P_X1 in the study 1-3  
# m_z: list of reported means of P_X2 in the study 1-9
# v_x: list of reported standard deviation of P_X1 in the study 1-3  
# v_z: list of reported standard deviation of P_X2 in the study 1-9
# rho: the correlation estimates of P_XZ calculated from IPD
# cutoff_bin: vector of cutoff values in the study 4-6 (ex. c(0.5,0,-0.5))
# cutoff_bin: matrix of cutoff values in the study 7-9 (ex. matrix(c(0,0.5,0.2,0.5,-0.5,0.5),ncol=2,byrow=T))
# n: list of sample size per category (ex. list(100,100,100,c(50,50),c(30,70),c(30,30,40),c(40,40,20)))

core = function(coef,S,m_x,m_z,v_x,v_z,rho,cutoff_bin,cutoff_ter,n){
	
	# Calculate the true parameter of distribution of X1
	par.X1 = optim(par = c(0, 1), fn = Likelihood(cutoff_bin,cutoff_ter,m_x,v_x,n))

	# Prepare lists to store the results	
	dist = list(9)

	# Construct the joint distribution of P_X1X2
	for (i in 1:9){
		if(i %in% c(1,2,3)){
			mu = c(m_x[[i]],m_z[[i]])
			Sigma_x = matrix(c(v_x[[i]]^2, rho*sqrt(v_x[[i]]^2*v_z[[i]]^2),
				rho*sqrt(v_x[[i]]^2*v_z[[i]]^2), v_z[[i]]^2), ncol = 2, nrow = 2)
			
			dist[[i]] = list(mu,Sigma_x)
		}else{
			mu = c(par.X1$par[1],m_z[[i]])
			Sigma_x = matrix(c(par.X1$par[2]^2, rho*sqrt(par.X1$par[2]^2*v_z[[i]]^2),
				rho*sqrt(par.X1$par[2]^2*v_z[[i]]^2), v_z[[i]]^2), ncol = 2, nrow = 2)
			
			dist[[i]] = list(mu,Sigma_x)
		}	
	}
	
	# To prepare initial value, which are the estimates from meta-analysis using first 3 studies
	beta_t = mvmeta(formula=rbind(coef[1,],coef[2,],coef[3,])~1,S=list(S[[1]],S[[2]],S[[3]]),method="reml")

	# calculate our proposed estimator
	propose = optim(
		c(coef(beta_t),summary(beta_t)$Psi[1,1],summary(beta_t)$Psi[2,2],summary(beta_t)$Psi[3,3],summary(beta_t)$Psi[1,2],summary(beta_t)$Psi[1,3],summary(beta_t)$Psi[2,3]),
		function(x){ll.meta(x,coef,dist,S)},method="BFGS", control=list(maxit=10000)
	)
	return(propose$par[1:3])
}
